/* BeaSTM Add-On (c) Thorbjörn Jemander */

#include "/boot/home/Development/Projects/BeaSTM/GUI/Windows/ImageWindow.h"
#include "/boot/home/Development/Projects/BeaSTM/Misc/PolygoneArea.h"

#define max(x,y) ((x)>(y)?(x):(y))
#define min(x,y) ((x)<(y)?(x):(y))


#pragma export on
extern "C" 
{
bool PreProcess( STMPolygoneArea ***argArea );
bool Process( uint16 *argSourceData, uint16 *argDestinationData, uint32 XSize, uint32 YSize );
}
#pragma export reset

BPoint YMin( BPoint argPoint1, BPoint argPoint2, BPoint argPoint3 );
BPoint YMid( BPoint argPoint1, BPoint argPoint2, BPoint argPoint3 );
BPoint YMax( BPoint argPoint1, BPoint argPoint2, BPoint argPoint3 );
BPoint XMin( BPoint argPoint1, BPoint argPoint2, BPoint argPoint3 );
BPoint XMid( BPoint argPoint1, BPoint argPoint2, BPoint argPoint3 );
BPoint XMax( BPoint argPoint1, BPoint argPoint2, BPoint argPoint3 );

double Calc_S1_ForTriangle( STMTriangle *argTriangle, uint16 *argSourceData, uint32 argXSize );
double Calc_S2_ForTriangle( STMTriangle *argTriangle, uint16 *argSourceData, uint32 argXSize );
double Calc_S3_ForTriangle( STMTriangle *argTriangle, uint16 *argSourceData, uint32 argXSize );
double Calc_A_ForTriangle( STMTriangle *argTriangle );
double Calc_B_ForTriangle( STMTriangle *argTriangle );
double Calc_C_ForTriangle( STMTriangle *argTriangle );
double Calc_D_ForTriangle( STMTriangle *argTriangle );
double Calc_E_ForTriangle( STMTriangle *argTriangle );
double Calc_F_ForTriangle( STMTriangle *argTriangle );

void	CopySubtractedPlane( uint16 *argDestData, uint16 *argSourceData, uint32 argXSize, uint32 argYSize, double a, double b, double c );

STMPolygoneArea *PlaneArea = NULL;

bool
Process( uint16 *argSourceData, uint16 *argDestinationData, uint32 argXSize, uint32 argYSize )
{
	
	uint32	i;
	
	double 	S1	= 0,
			S2	= 0,
			S3 	= 0,
			A	= 0,
			B	= 0,
			C	= 0,
			D	= 0,
			E	= 0,
			F	= 0,
			a,
			b,
			c;
	
	STMTriangle *Triangle;  
	
	for( i=0; i<PlaneArea->GetNumberOfTriangles(); i++ )
	{
		Triangle = PlaneArea->GetTriangle( i );
		S1 += Calc_S1_ForTriangle( Triangle, argSourceData, argXSize );
		S2 += Calc_S2_ForTriangle( Triangle, argSourceData, argXSize );
		S3 += Calc_S3_ForTriangle( Triangle, argSourceData, argXSize );
		A += Calc_A_ForTriangle( Triangle );
		B += Calc_B_ForTriangle( Triangle );
		C += Calc_C_ForTriangle( Triangle );
		D += Calc_D_ForTriangle( Triangle );
		E += Calc_E_ForTriangle( Triangle );
		F += Calc_F_ForTriangle( Triangle );
	}

	c = ( (B*A-C*C) * (S3*A-S1*D) - (E*A-D*C) * (S2*A-C*S1 ) ) / 
		( (F*A-D*D) * (B*A-C*C)   - (E*A-D*C) * (E*A-D*C ) );
	
	b = ( (S2*A-S1*C)-c*(E*A-D*C) ) / ( B*A-C*C );
	
	a = (S1-D*c-C*b)/A;

	CopySubtractedPlane( argDestinationData, argSourceData, argXSize, argYSize, a,b,c );

	return true;
}

bool
PreProcess( STMPolygoneArea ***argArea )
{
	*argArea = &PlaneArea;
	
	return true;
}

double
Calc_S1_ForTriangle( STMTriangle *argTriangle, uint16 *argSourceData, uint32 argXSize )
{
	double 	S1 = 0, S1_p = 0;

	double	i,j;

	double 	j_plus,
			j_minus,
			k_plus,
			k_minus,
			m_plus,
			m_minus;

	BPoint	Top, Middle, Bottom;
	
	
	Top 	= YMax( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Middle 	= YMid( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Bottom 	= YMin( argTriangle->point1, argTriangle->point2, argTriangle->point3 );

	if( Middle.x > Top.x )
	{
		k_plus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_minus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	else
	{
		k_minus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_plus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	
	m_plus = 0;
	m_minus = 0;
	
	for( i = Bottom.y; i < Middle.y; i++ )
	{
		j_minus = Bottom.x + k_minus * (i-Bottom.y );
		j_plus = Bottom.x + k_plus * (i-Bottom.y );
		
		S1_p = 0;
		
		for( j = j_minus; j <= j_plus; j++ )
			S1_p += argSourceData[ argXSize * (uint32)i + (uint32)j ];

		S1 += i*S1_p;

	}

	if( Middle.x > Top.x )
		k_plus = ( Top.x - Middle.x )/(Top.y-Middle.y);
	else
		k_minus = ( Top.x - Middle.x )/(Top.y-Middle.y);

	m_minus = j_minus;
	m_plus = j_plus;

	for( i= Middle.y; i<= Top.y; i++ )
	{
		j_minus = m_minus + k_minus * (i-Middle.y );
		j_plus = m_plus + k_plus * (i-Middle.y );

		S1_p = 0;

		for( j = j_minus; j <= j_plus; j++ )
			S1_p += argSourceData[ argXSize * (uint32)i + (uint32)j ];

		S1 += i*S1_p;
	}

	return S1;
}

double
Calc_S2_ForTriangle( STMTriangle *argTriangle, uint16 *argSourceData, uint32 argXSize  )
{
	double 	S2 = 0;

	double	i,j;

	double 	j_plus,
			j_minus,
			k_plus,
			k_minus,
			m_plus,
			m_minus;

	BPoint	Top, Middle, Bottom;
	
	Top 	= YMax( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Middle 	= YMid( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Bottom 	= YMin( argTriangle->point1, argTriangle->point2, argTriangle->point3 );

	if( Middle.x > Top.x )
	{
		k_plus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_minus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	else
	{
		k_minus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_plus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	
	m_plus = 0;
	m_minus = 0;
	
	for( i = Bottom.y; i < Middle.y; i++ )
	{
		j_minus = Bottom.x + k_minus * (i-Bottom.y );
		j_plus = Bottom.x + k_plus * (i-Bottom.y );
		
		for( j = j_minus; j <= j_plus; j++ )
			S2 += j*argSourceData[ argXSize * (uint32)i + (uint32)j ];
	}

	if( Middle.x > Top.x )
		k_plus = ( Top.x - Middle.x )/(Top.y-Middle.y);
	else
		k_minus = ( Top.x - Middle.x )/(Top.y-Middle.y);

	m_minus = j_minus;
	m_plus = j_plus;

	for( i= Middle.y; i<= Top.y; i++ )
	{
		j_minus = m_minus + k_minus * (i-Middle.y );
		j_plus = m_plus + k_plus * (i-Middle.y );

		for( j = j_minus; j <= j_plus; j++ )
			S2 += j*argSourceData[ argXSize * (uint32)i + (uint32)j ];
	}

	return S2;
}

double
Calc_S3_ForTriangle( STMTriangle *argTriangle, uint16 *argSourceData, uint32 argXSize  )
{
	double 	S3 = 0;

	double	i,j;

	double 	j_plus,
			j_minus,
			k_plus,
			k_minus,
			m_plus,
			m_minus;

	BPoint	Top, Middle, Bottom;
	
	Top 	= YMax( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Middle 	= YMid( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Bottom 	= YMin( argTriangle->point1, argTriangle->point2, argTriangle->point3 );

	if( Middle.x > Top.x )
	{
		k_plus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_minus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	else
	{
		k_minus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_plus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	
	m_plus = 0;
	m_minus = 0;
	
	for( i = Bottom.y; i < Middle.y; i++ )
	{
		j_minus = Bottom.x + k_minus * (i-Bottom.y );
		j_plus = Bottom.x + k_plus * (i-Bottom.y );
		
		for( j = j_minus; j <= j_plus; j++ )
			S3 += argSourceData[ argXSize * (uint32)i + (uint32)j ];
	}

	if( Middle.x > Top.x )
		k_plus = ( Top.x - Middle.x )/(Top.y-Middle.y);
	else
		k_minus = ( Top.x - Middle.x )/(Top.y-Middle.y);

	m_minus = j_minus;
	m_plus = j_plus;

	for( i= Middle.y; i<= Top.y; i++ )
	{
		j_minus = m_minus + k_minus * (i-Middle.y );
		j_plus = m_plus + k_plus * (i-Middle.y );

		for( j = j_minus; j <= j_plus; j++ )
			S3 += argSourceData[ argXSize * (uint32)i + (uint32)j ];
	}

	return S3;
}

double
Calc_A_ForTriangle( STMTriangle *argTriangle )
{
	double 	A = 0;

	double	i;

	double 	j_plus,
			j_minus,
			k_plus,
			k_minus,
			m_plus,
			m_minus;

	BPoint	Top, Middle, Bottom;
	

	Top 	= YMax( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Middle 	= YMid( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Bottom 	= YMin( argTriangle->point1, argTriangle->point2, argTriangle->point3 );


	if( Middle.x > Top.x )
	{
		k_plus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_minus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	else
	{
		k_minus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_plus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	
	m_plus = 0;
	m_minus = 0;
	
	for( i = Bottom.y; i < Middle.y; i++ )
	{
		j_minus = Bottom.x + k_minus * (i-Bottom.y );
		j_plus = Bottom.x + k_plus * (i-Bottom.y );

	
		A += i*i*(j_plus-j_minus+1);
	
	}

	if( Middle.x > Top.x )
		k_plus = ( Top.x - Middle.x )/(Top.y-Middle.y);
	else
		k_minus = ( Top.x - Middle.x )/(Top.y-Middle.y);

	m_minus = j_minus;
	m_plus = j_plus;

	for( i= Middle.y; i<= Top.y; i++ )
	{
		j_minus = m_minus + k_minus * (i-Middle.y );
		j_plus = m_plus + k_plus * (i-Middle.y );

		A += i*i*(j_plus-j_minus+1);
	}

	return A;
}

double
Calc_B_ForTriangle( STMTriangle *argTriangle )
{
	double 	B = 0;

	double	i;

	double 	j_plus,
			j_minus,
			k_plus,
			k_minus,
			m_plus,
			m_minus;

	BPoint	Top, Middle, Bottom;
	
	Top 	= YMax( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Middle 	= YMid( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Bottom 	= YMin( argTriangle->point1, argTriangle->point2, argTriangle->point3 );

	if( Middle.x > Top.x )
	{
		k_plus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_minus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	else
	{
		k_minus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_plus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	
	m_plus = 0;
	m_minus = 0;
	
	for( i = Bottom.y; i < Middle.y; i++ )
	{
		j_minus = Bottom.x + k_minus * (i-Bottom.y );
		j_plus = Bottom.x + k_plus * (i-Bottom.y );
		B += j_plus*j_plus*j_plus/3+j_plus*j_plus/2+j_plus/6-((j_minus-1)*(j_minus-1)*(j_minus-1)/3+(j_minus-1)*(j_minus-1)/2+(j_minus-1)/6);
	}

	if( Middle.x > Top.x )
		k_plus = ( Top.x - Middle.x )/(Top.y-Middle.y);
	else
		k_minus = ( Top.x - Middle.x )/(Top.y-Middle.y);

	m_minus = j_minus;
	m_plus = j_plus;

	for( i= Middle.y; i<= Top.y; i++ )
	{
		j_minus = m_minus + k_minus * (i-Middle.y );
		j_plus = m_plus + k_plus * (i-Middle.y );

		B += j_plus*j_plus*j_plus/3+j_plus*j_plus/2+j_plus/6-((j_minus-1)*(j_minus-1)*(j_minus-1)/3+(j_minus-1)*(j_minus-1)/2+(j_minus-1)/6);
	}

	return B;
}

double
Calc_C_ForTriangle( STMTriangle *argTriangle )
{
	double 	C = 0;

	double	i;

	double 	j_plus,
			j_minus,
			k_plus,
			k_minus,
			m_plus,
			m_minus;

	BPoint	Top, Middle, Bottom;
	
	Top 	= YMax( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Middle 	= YMid( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Bottom 	= YMin( argTriangle->point1, argTriangle->point2, argTriangle->point3 );

	if( Middle.x > Top.x )
	{
		k_plus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_minus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	else
	{
		k_minus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_plus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	
	m_plus = 0;
	m_minus = 0;
	
	for( i = Bottom.y; i < Middle.y; i++ )
	{
		j_minus = Bottom.x + k_minus * (i-Bottom.y );
		j_plus = Bottom.x + k_plus * (i-Bottom.y );
		C += 0.5*i*(j_plus*(j_plus+1)-(j_minus-1)*j_minus);
	}

	if( Middle.x > Top.x )
		k_plus = ( Top.x - Middle.x )/(Top.y-Middle.y);
	else
		k_minus = ( Top.x - Middle.x )/(Top.y-Middle.y);

	m_minus = j_minus;
	m_plus = j_plus;

	for( i= Middle.y; i<= Top.y; i++ )
	{
		j_minus = m_minus + k_minus * (i-Middle.y );
		j_plus = m_plus + k_plus * (i-Middle.y );

		C += 0.5*i*(j_plus*(j_plus+1)-(j_minus-1)*j_minus);
	}

	return C;

}

double
Calc_D_ForTriangle( STMTriangle *argTriangle )
{
	double 	D = 0;

	double	i;

	double 	j_plus,
			j_minus,
			k_plus,
			k_minus,
			m_plus,
			m_minus;

	BPoint	Top, Middle, Bottom;
	
	Top 	= YMax( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Middle 	= YMid( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Bottom 	= YMin( argTriangle->point1, argTriangle->point2, argTriangle->point3 );

	if( Middle.x > Top.x )
	{
		k_plus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_minus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	else
	{
		k_minus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_plus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	
	m_plus = 0;
	m_minus = 0;
	
	for( i = Bottom.y; i < Middle.y; i++ )
	{
		j_minus = Bottom.x + k_minus * (i-Bottom.y );
		j_plus = Bottom.x + k_plus * (i-Bottom.y );
		D += i*(j_plus-j_minus+1);
	}

	if( Middle.x > Top.x )
		k_plus = ( Top.x - Middle.x )/(Top.y-Middle.y);
	else
		k_minus = ( Top.x - Middle.x )/(Top.y-Middle.y);

	m_minus = j_minus;
	m_plus = j_plus;

	for( i= Middle.y; i<= Top.y; i++ )
	{
		j_minus = m_minus + k_minus * (i-Middle.y );
		j_plus = m_plus + k_plus * (i-Middle.y );

		D += i*(j_plus-j_minus+1);
	}

	return D;
}

double
Calc_E_ForTriangle( STMTriangle *argTriangle )
{
	double 	E = 0;

	double	i;

	double 	j_plus,
			j_minus,
			k_plus,
			k_minus,
			m_plus,
			m_minus;

	BPoint	Top, Middle, Bottom;
	
	Top 	= YMax( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Middle 	= YMid( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Bottom 	= YMin( argTriangle->point1, argTriangle->point2, argTriangle->point3 );

	if( Middle.x > Top.x )
	{
		k_plus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_minus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	else
	{
		k_minus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_plus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	
	
	m_plus = 0;
	m_minus = 0;
	
	for( i = Bottom.y; i < Middle.y; i++ )
	{
		j_minus = Bottom.x + k_minus * (i-Bottom.y );
		j_plus = Bottom.x + k_plus * (i-Bottom.y );
		E +=  0.5*(j_plus*(j_plus+1)-(j_minus-1)*j_minus);
	}

	if( Middle.x > Top.x )
		k_plus = ( Top.x - Middle.x )/(Top.y-Middle.y);
	else
		k_minus = ( Top.x - Middle.x )/(Top.y-Middle.y);

	m_minus = j_minus;
	m_plus = j_plus;

	for( i= Middle.y; i<= Top.y; i++ )
	{
		j_minus = m_minus + k_minus * (i-Middle.y );
		j_plus = m_plus + k_plus * (i-Middle.y );

		E +=  0.5*(j_plus*(j_plus+1)-(j_minus-1)*j_minus);
	}

	return E;
}

double
Calc_F_ForTriangle( STMTriangle *argTriangle )
{
	double 	F = 0;

	double	i;

	double 	j_plus,
			j_minus,
			k_plus,
			k_minus,
			m_plus,
			m_minus;

	BPoint	Top, Middle, Bottom;
	
	Top 	= YMax( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Middle 	= YMid( argTriangle->point1, argTriangle->point2, argTriangle->point3 );
	Bottom 	= YMin( argTriangle->point1, argTriangle->point2, argTriangle->point3 );

	if( Middle.x > Top.x )
	{
		k_plus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_minus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	else
	{
		k_minus = (Middle.x-Bottom.x)/(Middle.y-Bottom.y);
		k_plus = (Top.x-Bottom.x)/(Top.y-Bottom.y);
	}
	
	m_plus = 0;
	m_minus = 0;
	
	for( i = Bottom.y; i < Middle.y; i++ )
	{
		j_minus = Bottom.x + k_minus * (i-Bottom.y );
		j_plus = Bottom.x + k_plus * (i-Bottom.y );
		F += (j_plus-j_minus+1);
	}

	if( Middle.x > Top.x )
		k_plus = ( Top.x - Middle.x )/(Top.y-Middle.y);
	else
		k_minus = ( Top.x - Middle.x )/(Top.y-Middle.y);

	m_minus = j_minus;
	m_plus = j_plus;


	for( i= Middle.y; i<= Top.y; i++ )
	{
		j_minus = m_minus + k_minus * (i-Middle.y );
		j_plus = m_plus + k_plus * (i-Middle.y );

		F += (j_plus-j_minus+1);
	}

	return F;
}

void
CopySubtractedPlane( uint16 *argDestData, uint16 *argSourceData, uint32 argXSize, uint32 argYSize, double a, double b, double c )
{
	int32 x,y;

	for( x=0; x<argXSize; x++ )
		for( y=0; y<argYSize; y++ )
			argDestData[ argXSize*y+x ] = 32768 + argSourceData[argXSize*y+x] - (uint16)(a*y + b*x + c);
}


BPoint
YMax( BPoint argPoint1, BPoint argPoint2, BPoint argPoint3 )
{
	if( argPoint1.y > argPoint2.y )
	{
		if( argPoint1.y > argPoint3.y )
			return argPoint1;
		else
			return argPoint3;
	}
	else
		if( argPoint2.y > argPoint3.y )
			 return argPoint2;
		else
			return argPoint3; 
}

BPoint
YMid( BPoint argPoint1, BPoint argPoint2, BPoint argPoint3 )
{
	if( argPoint1.y < argPoint2.y && argPoint1.y > argPoint3.y  || argPoint1.y < argPoint3.y && argPoint1.y > argPoint2.y  )
		return argPoint1;

	if( argPoint2.y < argPoint1.y && argPoint2.y > argPoint3.y  || argPoint2.y < argPoint3.y && argPoint2.y > argPoint1.y  )
		return argPoint2;

	if( argPoint3.y < argPoint2.y && argPoint3.y > argPoint1.y  || argPoint3.y < argPoint1.y && argPoint3.y > argPoint2.y  )
		return argPoint3;

	return argPoint1;
}

BPoint
YMin( BPoint argPoint1, BPoint argPoint2, BPoint argPoint3 )
{
	if( argPoint1.y < argPoint2.y )
	{
		if( argPoint1.y < argPoint3.y )
			return argPoint1;
		else
			return argPoint3;
	}
	else
		if( argPoint2.y < argPoint3.y )
			 return argPoint2;
		else
			return argPoint3; 
}

