
/* Device driver for the KEITHLY 575 
   Copyright (c) 1997 Thorbjörn Jemander */

#ifndef K575_H
#define K575_H

#include <Drivers.h>
#include <drivers/KernelExport.h>
#include "Main/GlobalDefs.h"
#include "Parameters/ADDAParams.h"
#include "Parameters/ScanningParameters.h"
#include "Parameters/TimingParams.h"
#include "Parameters/ImageParameters.h"
#include "GUI/Windows/ImageWindow.h"
#include "GUI/Views/ImageView.h"
#include "Image/Image.h"

#define STM_DRIVER_DEBUG 2

#ifdef STM_DRIVER_DEBUG
#define	ENABLE_DEBUG 	set_dprintf_enabled( true );
#define D(x) dprintf(x) 
#else
#define D(x)
#endif

#define CMDA1						0x00
#define CMDB1						0x01
#define CMDC1						0x1A
#define CMDD1						0x1B
#define CMDA2						0x02
#define CMDB2						0x03
#define CMDC2						0x18
#define CMDA3						0x04
#define CMDB3						0x05
#define CMDC3						0x19
#define CMDA4						0x06
#define CMDB4						0x07
#define STROBE						0x1D
#define CMDA5						0x08
#define CMDB5						0x09
#define CMDA6						0x0A
#define CMDB6						0x0B
#define CMDA7						0x0C
#define CMDB7						0x0D
#define CMDA8						0x0E
#define CMDB8						0x0F
#define CMDA9						0x10
#define CMDB9						0x11
#define CMDA10						0x12
#define CMDB10						0x13
#define RW_COUNTER_0				0x40
#define RW_COUNTER_1				0x41
#define RW_COUNTER_2				0x42
#define COUNTER_CONTROL 			0x43
#define TIMER_GLOBAL				0x60
#define TIMER_STATUS				0x61
#define CLEAR_INTERRUPT				0x62
#define SET_INT_LEVEL				0x63
#define BOX_ID						0x1C

#define UP 		true
#define DOWN	false

enum
{	STM_PREPARE_SCAN= B_DEVICE_OP_CODES_END +1,
	STM_BEGIN_SCAN,
	STM_BEGIN_SIMULATED_SCAN,
	STM_STOP_SCANNING,
	STM_FINISH_THIS_SCAN,
	STM_SEND_PULSES,
	STM_ALLOCATE_KERNEL_AREA,
	STM_DELETE_KERNEL_AREA,
	STM_SET_BIAS_AND_CURRENT
};
void		MoveTipToOrigo();
void		MoveTipToDAPoint( uint32 XPosition, uint32 YPosition );

int16		TakeSamplesFromADC( int32		argNumberOfSamplesToTake );
status_t	ResetAD();
status_t	ConfigADCAndDACFromParameters( InternalADDA *Params );
status_t	InitExternalDAC();
status_t	InitADCAndDAC();
status_t	SetDARange( int32	argChannel, int32 argRange );
status_t	SetDAValue(int32	argChannel, int32 argValue );
inline 
uint16		CalcNewIntegerBias( float argBias );
inline 
uint16		CalcNewIntegerCurrent( float argCurrent );
inline
void		SetBiasAndCurrent( float argBias, float argCurrent );
inline
void		SetIntegerBiasAndCurrent( uint16 argBias, uint16 argCurrent );
inline
void		SetXDACStepDirectionTo( bool direction );
inline
void		SetYDACStepDirectionTo( bool direction );
inline
void		StepOneXPixel(	int32 argNumberOfDASteps, int32 argDelayBetweenSteps );
inline
void		StepOneYPixel(	int32 argNumberOfDASteps, int32 argDelayBetweenSteps );
inline 
void		microsecond_delay( int64 x );
inline
void		KWrite( 			uint8 value,			int command_register );
inline
uint8		KRead( 				int command_register );

int32		CreateArea(			void *data);
int32		DeleteArea(			void *data);

void		DoPreScans( 		uint32 argNumberOfPrescans );
int32		RealScan(			void *);
int32		SimulatedScan(		void *);
status_t	K575_open( 			const char *name, uint32 flags, void **cookie );
status_t	K575_close( 		void *cookie );
static
status_t	K575_free_cookie(	void *cookie );
status_t	K575_control(		void * cookie, uint32 op, void *data, size_t len);
ssize_t		K575_read(			void * cookie, off_t position, void *data, size_t *numBytes);
ssize_t		K575_write(			void * cookie, off_t position,const void *data, size_t *numBytes);

#endif
